#AutoIt3Wrapper_AU3Check_Parameters=-w 3 -w 4 -w 5 -w 6 -d

#include <Constants.au3>
#include <Array.au3>
#include <InetConstants.au3>
#include "HttpApi.au3"

#cs
	HTTPAPI.au3  - An AutoIt implementation of Microsoft's HTTP Server API functionality.

	You can read more about Microsoft's HTTP Server API starting here:
	https://docs.microsoft.com/en-us/windows/win32/http/http-api-start-page

	This script shows you a granular example of how to set up a simple HTTP Server that
	listens from several different GET requests.  It also shows very basic POST request
	processing by echoing some of the POST request information.  Of course you can have
	it listen for any type of HTTP request (GET, POST, DELETE, etc) that you would like.
	GET processing can be tested using any browser.  To test POSTs and other types of
	requests, you would need a tool that allows you to test RESTful APIs like POSTMAN (or
	RESTer which is my preferred Firefox plugin for REST API testing).

	To see the functionality in this script, simply run the script.  Upon launch, it should show
	you a message box with the registered test URL which is ""http://127.0.0.1:9000"/a3server"
	by default.  You can change the default by modifying the global constants below.

	NOTE: The HTTP Server script will continue to wait for and process requests until it is shut down.
	      See the URL path below to cleanly shut down the HTTP Server script.

	The following URLS are set up to be recognized by this script:

	http://127.0.0.1:9000/a3server            - Basically a Hello response from the HTTP Server
	http://127.0.0.1:9000/a3server/test1      - A vanilla response showing the request was recognized.
	http://127.0.0.1:9000/a3server/myip       - This request will display your current public IP address.
	http://127.0.0.1:9000/a3server/ipconfig   - This request will display the output of an IPCONFIG command.
	http://127.0.0.1:9000/a3server/getheaders - This request will display the HTTP request's headers
	http://127.0.0.1:9000/a3server/redirect1  - This request will redirect to www.google.com
	http://127.0.0.1:9000/a3server/stop       - CLEANLY SHUTS DOWN THE HTTP SERVER AND DISPLAYS A MESSAGE.

	Any other GET paths that start with "http://localhost:9000/a3server" will get a 404 NOT FOUND
	response.  Any other HTTP verbs like DELETE, will receive a 503 response.  All of the specific
	responses were scripted in the example below.  Most of the "magic" occurs in the following
	functions:

	process_requests()
	process_get_request()
	process_post_request()

	If you are the kind of person that likes reading the last chapter of a novel first, that's
	probably where you will like to start.  :)

	Enjoy,
	TheXman
#ce

;Global constants
Const  $HTTP_HOST          = "http://127.0.0.1:9000", _
       $HTTP_PATH          = "/a3server"

;Global integers
Global $giServerSessionId = 0, _
       $giUrlGroupId      = 0

;Global handles
Global $ghRequestQueue    = -1

;Global arrays
Global $gaUrls[0]

main()

;==========================================================================

Func main()

	httpapi_udf_startup()
	httpapi_initialize()
	create_server_session()
	create_url_group()
	create_request_queue()
	bind_request_queue_to_url_group()
	add_url_to_url_group($HTTP_HOST & $HTTP_PATH)

	MsgBox($MB_ICONINFORMATION, "INFO", _
		   "HTTP Server is now listening at URL:" & @CRLF & @CRLF & $HTTP_HOST & $HTTP_PATH)

	ShellExecute($HTTP_HOST & $HTTP_PATH)

	process_requests()

EndFunc

Func httpapi_udf_startup($bDebug = False)
	;Initialize UDF
	_HTTPAPI_Startup($bDebug)
	If @error Then
		MsgBox($MB_ICONERROR, "ERROR", _
			"_HTTPAPI_Startup failed with @error =" & @error & " @extended = " & @extended & @CRLF & _
			_HTTPAPI_GetLastErrorMessage())
		Exit 1
	EndIf
EndFunc

Func httpapi_initialize()
	;Intialize the API
	_HTTPAPI_HttpInitialize()
	If @error Then
		MsgBox($MB_ICONERROR, "_HTTPAPI_HttpInitialize ERROR", _
			"@error =" & @error & "  @extended = " & @extended & @CRLF & @CRLF & _
			_HTTPAPI_GetLastErrorMessage())
		Exit 1
	EndIf
	OnAutoItExitRegister(httpapi_terminate)
EndFunc

Func httpapi_terminate()
	;Terminate the API
	_HTTPAPI_HttpTerminate()
	If @error Then
		MsgBox($MB_ICONERROR, "_HTTPAPI_HttpTerminate ERROR", _
			"@error =" & @error & "  @extended = " & @extended & @CRLF & @CRLF & _
			_HTTPAPI_GetLastErrorMessage())
	EndIf

	MsgBox($MB_ICONINFORMATION, "INFO", "HTTP Server has successfully terminated.")
EndFunc

Func create_server_session()
	;Create a server session
	$giServerSessionId = _HTTPAPI_HttpCreateServerSession()
	If @error Then
		MsgBox($MB_ICONERROR, "_HTTPAPI_HttpCreateServerSession ERROR", _
			"@error =" & @error & "  @extended = " & @extended & @CRLF & @CRLF & _
			_HTTPAPI_GetLastErrorMessage())
		Exit 1
	EndIf
	OnAutoItExitRegister(close_server_session)
EndFunc

Func close_server_session()
	;Close server session
	_HTTPAPI_HttpCloseServerSession($giServerSessionId)
	If @error Then
		MsgBox($MB_ICONERROR, "_HTTPAPI_HttpCloseServerSession ERROR", _
			"@error =" & @error & "  @extended = " & @extended & @CRLF & @CRLF & _
			_HTTPAPI_GetLastErrorMessage())
	EndIf
EndFunc

Func create_url_group()
	;Create a URL group
	$giUrlGroupId = _HTTPAPI_HttpCreateUrlGroup($giServerSessionId)
	If @error Then
		MsgBox($MB_ICONERROR, "_HTTPAPI_HttpCreateUrlGroup ERROR", _
			"@error =" & @error & "  @extended = " & @extended & @CRLF & @CRLF & _
			_HTTPAPI_GetLastErrorMessage())
		Exit 1
	EndIf
	OnAutoItExitRegister(close_url_group)
EndFunc

Func close_url_group()
	;Close server session
	_HTTPAPI_HttpCloseUrlGroup($giUrlGroupId)
	If @error Then
		MsgBox($MB_ICONERROR, "_HTTPAPI_HttpCloseUrlGroup ERROR", _
			"@error =" & @error & "  @extended = " & @extended & @CRLF & @CRLF & _
			_HTTPAPI_GetLastErrorMessage())
	EndIf
EndFunc

Func create_request_queue()
	;Create a new request queue
	$ghRequestQueue = _HTTPAPI_HttpCreateRequestQueue()
	If @error Then
		MsgBox($MB_ICONERROR, "_HTTPAPI_HttpCreateRequestQueue ERROR", _
			"@error =" & @error & "  @extended = " & @extended & @CRLF & @CRLF & _
			_HTTPAPI_GetLastErrorMessage())
		Exit 1
	EndIf
	OnAutoItExitRegister(close_request_queue)
EndFunc

Func close_request_queue()
	;Shut down request queue
	_HTTPAPI_HttpShutdownRequestQueue($ghRequestQueue)
	If @error Then
		MsgBox($MB_ICONERROR, "_HTTPAPI_HttpShutdownRequestQueue ERROR", _
			"@error =" & @error & "  @extended = " & @extended & @CRLF & @CRLF & _
			_HTTPAPI_GetLastErrorMessage())
	EndIf
EndFunc

Func bind_request_queue_to_url_group()
	Local $tBindingInfo = DllStructCreate($__HTTPAPI_gtagHTTP_BINDING_INFO)

	;Set Binding Info
	$tBindingInfo.Flags              = $__HTTPAPI_HTTP_PROPERTY_FLAG_PRESENT
	$tBindingInfo.RequestQueueHandle = $ghRequestQueue

	;Bind request
	_HTTPAPI_HttpSetUrlGroupProperty( _
	                                 $giUrlGroupId, _
									 $__HTPPAPI_HttpServerBindingProperty, _
									 DllStructGetPtr($tBindingInfo), _
									 DllStructGetSize($tBindingInfo))
	If @error Then
		MsgBox($MB_ICONERROR, "_HTTPAPI_HttpSetUrlGroupProperty ERROR", _
			"@error =" & @error & "  @extended = " & @extended & @CRLF & @CRLF & _
			_HTTPAPI_GetLastErrorMessage())
		Exit 1
	EndIf
	OnAutoItExitRegister(unbind_request_queue_from_url_group)
EndFunc

Func unbind_request_queue_from_url_group()
	Local $tBindingInfo = DllStructCreate($__HTTPAPI_gtagHTTP_BINDING_INFO)

	;Set Binding Info
	$tBindingInfo.Flags              = 0
	$tBindingInfo.RequestQueueHandle = Null

	;Bind request
	_HTTPAPI_HttpSetUrlGroupProperty( _
	                                 $giUrlGroupId, _
									 $__HTPPAPI_HttpServerBindingProperty, _
									 DllStructGetPtr($tBindingInfo), _
									 DllStructGetSize($tBindingInfo))
	If @error Then
		MsgBox($MB_ICONERROR, "_HTTPAPI_HttpSetUrlGroupProperty ERROR", _
			"@error =" & @error & "  @extended = " & @extended & @CRLF & @CRLF & _
			_HTTPAPI_GetLastErrorMessage())
	EndIf
EndFunc

Func add_url_to_url_group($sUrl)
	;Add url to url group
	_HTTPAPI_HttpAddUrlToUrlGroup($giUrlGroupId, $sUrl)
	If @error Then
		MsgBox($MB_ICONERROR, "_HttpAddUrlToUrlGroup ERROR", _
			"@error =" & @error & "  @extended = " & @extended & @CRLF & @CRLF & _
			_HTTPAPI_GetLastErrorMessage())
		Exit 1
	EndIf

	;Store URL
	_ArrayAdd($gaUrls, $sUrl)

	OnAutoItExitRegister(remove_all_urls_from_url_group)
EndFunc

Func remove_all_urls_from_url_group()
	;Remove all URLS from URL Group
	_HTTPAPI_HttpRemoveUrlFromUrlGroup($giUrlGroupId, "", $__HTTPAPI_HTTP_URL_FLAG_REMOVE_ALL)
	If @error Then
		MsgBox($MB_ICONERROR, "_HTTPAPI_HttpSetUrlGroupProperty ERROR", _
			"@error =" & @error & "  @extended = " & @extended & @CRLF & @CRLF & _
			_HTTPAPI_GetLastErrorMessage())
	EndIf
EndFunc

Func cmd_capture($sCmd, $sWorkingDir = @WorkingDir)
	Local $iPID     = 0
	Local $sOutput  = ""

	;resolve defaults
	If $sWorkingDir = Default Then $sWorkingDir = @WorkingDir

	;execute command
	$iPID = Run(@ComSpec & ' /c ' & $sCmd, $sWorkingDir, @SW_HIDE, $STDERR_MERGED)
	If @error Then
		ConsoleWrite("Error: Unable to execute command - " & $sCmd & @CRLF)
		Return SetError(1, 0, "")
	EndIf

	;wait for process to end
	ProcessWaitClose($iPID)

	;get output
	$sOutput = StdoutRead($iPID)

	;if no output, set @error and return empty string
	If $sOutput = "" Then Return SetError(2, 0, "")

	;return output
	Return $sOutput
EndFunc

Func process_requests()
	Local $tBuffer = "", $tRequest = ""  ;, $tDataChunk = "", $tHeader = ""
	Local $sPath = ""

	;Wait for requests
	While 1
		;Receive request
		$tBuffer = _HTTPAPI_HttpReceiveHttpRequest($ghRequestQueue)
		If @error Then
			MsgBox($MB_ICONERROR, "_HTTPAPI_HttpReceiveHttpRequest ERROR", _
				"@error =" & @error & "  @extended = " & @extended & @CRLF & @CRLF & _
				_HTTPAPI_GetLastErrorMessage())
			Exit 1
		EndIf

		;Create structs for request and body
		$tRequest = DllStructCreate($__HTTPAPI_gtagHTTP_REQUEST_V2 & StringFormat("byte body[%i];", $__HTTPAPI_REQUEST_BODY_SIZE), DllStructGetPtr($tBuffer))

		;If stop requested
		$sPath =  _WinAPI_GetString($tRequest.pAbsPath)
		If StringRight($sPath, 1) = "/" Then $sPath = StringTrimRight($sPath, 1)
		If $sPath = $HTTP_PATH & "/stop" Then ExitLoop

		;Process request based on verb
		Switch $tRequest.Verb

			Case $__HTTPAPI_HttpVerbGET
				process_get_request(DllStructGetPtr($tRequest))

			Case $__HTTPAPI_HttpVerbPOST
				process_post_request(DllStructGetPtr($tRequest))

			Case Else
				_HTTPAPI_HttpSendHttpResponse($ghRequestQueue, $tRequest.RequestId, 503, "Not Implemented", "")
				If @error Then
					MsgBox($MB_ICONERROR, "_HTTPAPI_HttpSendHttpResponse ERROR", _
						"@error =" & @error & "  @extended = " & @extended & @CRLF & @CRLF & _
						_HTTPAPI_GetLastErrorMessage())
					Exit 1
				EndIf

		EndSwitch

		;Delete request resources
		$tBuffer = 0

		Sleep(100)
	WEnd
EndFunc

Func process_get_request($pRequest)
	Local $tRequest = ""
	Local $sPath = "", $sMsg = "", $sReason, $sCmdOutput = "", $sContentType = "text/html"
	Local $iStatusCode = 0
	Local $aReqHeaders[0][2]

	;Create structs for request and body
	$tRequest = DllStructCreate($__HTTPAPI_gtagHTTP_REQUEST_V2 & StringFormat("byte body[%i];", $__HTTPAPI_REQUEST_BODY_SIZE), $pRequest)

	;Send response based on requested URL path (i.e. /a3server/ipconfig)
	$sPath = _WinAPI_GetString($tRequest.pAbsPath)
	If StringRight($sPath, 1) = "/" Then $sPath = StringTrimRight($sPath, 1) ;Remove trailing "/" (if exists)

	Switch $sPath
		;/
		Case $HTTP_PATH
			$iStatusCode = 200
			$sReason     = "OK"
			$sMsg        = "<h3>GET request received!</h3>Hello from the AutoIt HTTP Server." & _
			               "<pre>" & @CRLF & @CRLF & _
	                       "The following URLS are set up to be recognized by this example script.  Give them a try." & @CRLF & @CRLF & _
	                       "http://127.0.0.1:9000/a3server            - Basically a Hello response from the HTTP Server" & @CRLF & _
	                       "http://127.0.0.1:9000/a3server/test1      - A vanilla response showing the request was recognized." & @CRLF & _
	                       "http://127.0.0.1:9000/a3server/myip       - This request will display your current public IP address." & @CRLF & _
	                       "http://127.0.0.1:9000/a3server/json       - This request will return a sample json response, with content-type set to application/json" & @CRLF & _
	                       "http://127.0.0.1:9000/a3server/ipconfig   - This request will display the output of an IPCONFIG command." & @CRLF & _
	                       "http://127.0.0.1:9000/a3server/getheaders - This request will display the HTTP request's headers" & @CRLF & _
	                       "http://127.0.0.1:9000/a3server/redirect1  - This request will redirect to www.google.com" & @CRLF & _
	                       "http://127.0.0.1:9000/a3server/stop       - Cleanly shuts down the HTTP server and display a message.</pre>"

		;/test1
		Case $HTTP_PATH & "/test1"
			$iStatusCode = 200
			$sReason     = "OK"
			$sMsg        = "<h3>GET request received!</h3>" & _
			               "Test1 request received."

		;/ipconfig
		Case $HTTP_PATH & "/ipconfig"
			$iStatusCode = 200
			$sReason     = "OK"

			;Run command, capture output, and send output back as response
			$sCmdOutput = cmd_capture("ipconfig /all")
			If @error Then
				$sMsg = "<h3>An error occured!</h3>@error = " & @error
			Else
				$sMsg = "<h3>Output</h3>" & _
				"<pre>" & $sCmdOutput & "</pre>"
			EndIf

		;/myip
		Case $HTTP_PATH & "/myip"
			$iStatusCode = 200
			$sReason     = "OK"

			;Run command, capture output, and send output back as response
			$sCmdOutput = InetRead("http://api.ipify.org", $INET_FORCEBYPASS + $INET_FORCERELOAD)
			If @error Then
				$sMsg = "<h3>An error occured!</h3>@error = " & @error
			Else
				$sMsg = "<h3>Output</h3>" & _
				"<pre>" & "My current public IP address is: " & BinaryToString($sCmdOutput, $SB_UTF8) & "</pre>"
			EndIf

		;/json
		Case $HTTP_PATH & "/json"
			$iStatusCode  = 200
			$sReason      = "OK"
			$sMsg         = '{"fld1": "Test", "fld2": false, "fld3": 12.5}'
			$sContentType = "application/json"

		;/getheaders
		Case $HTTP_PATH & "/getheaders"
			$iStatusCode = 200
			$sReason     = "OK"

			;Run command, capture output, and send output back as response
			$aReqHeaders = _HTTPAPI_GetRequestHeaders(DllStructGetPtr($tRequest, "UnknownHeaderCount"))
			$sCmdOutput  = ""
			For $i = 0 To UBound($aReqHeaders) - 1
				$sCmdOutput &= StringFormat("%s: %s\r\n", $aReqHeaders[$i][0], $aReqHeaders[$i][1])
			Next
			$sMsg = "<h3>Request Headers</h3>" & _
			        "<pre>" & $sCmdOutput & "</pre>"

		;/redirect1
		Case $HTTP_PATH & "/redirect1"
			$iStatusCode = 302
			$sReason     = "https://www.google.com"

		;Unhandled path
		Case Else
			$iStatusCode = 404
			$sReason     = "Not Found"
			$sMsg        = "<h2>Not Found</h2>" & _
			               "<hr>" & _
			               "<p>HTTP Error 404. The requested resource is not found.</p>"
	EndSwitch

	_HTTPAPI_HttpSendHttpResponse($ghRequestQueue, $tRequest.RequestId, $iStatusCode, $sReason, $sMsg, $sContentType)
	If @error Then
		MsgBox($MB_ICONERROR, "_HTTPAPI_HttpSendHttpResponse ERROR", _
			"@error =" & @error & "  @extended = " & @extended & @CRLF & @CRLF & _
			_HTTPAPI_GetLastErrorMessage())
		Exit 1
	EndIf
EndFunc

Func process_post_request($pRequest)
	Local $tRequest = ""
	Local $aHeaders
	Local $sResponse = ""

	;Create struct that points to the request and body
	$tRequest = DllStructCreate($__HTTPAPI_gtagHTTP_REQUEST_V2 & StringFormat("byte body[%i];", $__HTTPAPI_REQUEST_BODY_SIZE), $pRequest)

	;Build response
	$sResponse &= "POST Request Info:" & @CRLF
	$sResponse &= "Request struct size = " & DllStructGetSize($tRequest) & @CRLF
	$sResponse &= "Bytes received      = " & $tRequest.BytesReceived & @CRLF
	$sResponse &= "Full URL            = " & _WinAPI_GetString($tRequest.pFullUrl) & @CRLF
	$sResponse &= "Host                = " & StringRegExpReplace(_WinAPI_GetString($tRequest.pHost), "^([^/]+).*", "\1") & @CRLF
	$sResponse &= "AbsPath             = " & _WinAPI_GetString($tRequest.pAbsPath) & @CRLF
	$sResponse &= "QueryString         = " & _WinAPI_GetString($tRequest.pQueryString) & @CRLF
	$sResponse &= "Verb                = " & $tRequest.verb & @CRLF
	$sResponse &= "EntityChunkCount    = " & $tRequest.EntityChunkCount & @CRLF

	$aHeaders = _HTTPAPI_GetRequestHeaders(DllStructGetPtr($tRequest, "UnknownHeaderCount"))
	If IsArray($aHeaders) Then
		$sResponse &= @CRLF & "Request Headers:" & @CRLF
		For $i = 0 To UBound($aHeaders) - 1
			$sResponse &= StringFormat("%s: %s", $aHeaders[$i][0], $aHeaders[$i][1]) & @CRLF
		Next
	EndIf


    $sResponse &= @CRLF & "Request Body:" & @CRLF
	If $tRequest.pEntityChunks > 0 Then
        $sResponse &= _HTTPAPI_GetRequestBody($tRequest) & @CRLF & @CRLF
	EndIf
    If BitAND($tRequest.Flags, $__HTTPAPI_HTTP_REQUEST_FLAG_MORE_ENTITY_BODY_EXISTS) Then
        $sResponse &= _HTTPAPI_HttpReceiveRequestEntityBody($ghRequestQueue, $tRequest.RequestId) & @CRLF & @CRLF
    EndIf

	;Send response
	_HTTPAPI_HttpSendHttpResponse($ghRequestQueue, $tRequest.RequestId, _
	                              200, "OK", _
								  "Hello from the AutoIt HTTP Server.  POST request received!" & _
								  "<br><br>" & _
								  "<pre>" & $sResponse & "</pre>")
	If @error Then
		MsgBox($MB_ICONERROR, "_HTTPAPI_HttpSendHttpResponse ERROR", _
			"@error =" & @error & "  @extended = " & @extended & @CRLF & @CRLF & _
			_HTTPAPI_GetLastErrorMessage())
		Exit 1
	EndIf
EndFunc
